import datetime
import os
from pathlib import Path

import click
from great_expectations import DataContext
from great_expectations.cli import toolkit
from great_expectations.cli.toolkit import create_empty_suite
from kedro.framework.context import KedroContext, load_context

from ..data import generate_datasource_name


@click.command(name="suites")
@click.option(
    "--directory",
    "-d",
    default=None,
    help="The project's great_expectations directory.",
)
@click.option("--empty", "empty", flag_value=True, help="Creates empty suites.")
@click.option(
    "--replace", "replace", flag_value=True, help="Replace any existing suites."
)
@click.option(
    "--batch-kwargs",
    default=None,
    help="Additional keyword arguments to be provided to get_batch when loading the data asset. Must be a valid JSON dictionary",
)
def suite_new(directory, empty, replace, batch_kwargs):
    """
    Create Great Expectation Suites based on the kedro catalog using the BasicSuiteBuilderProfiler.

    If you wish to create suites without using the BasicSuiteBuilderProfiler, add the `--empty` flag.
    """

    kedro_context = load_context(Path.cwd())
    ge_context = toolkit.load_data_context_with_error_handling(directory)
    generate_basic_suites(kedro_context, ge_context, empty, replace, batch_kwargs)


def generate_basic_suite_name(dataset_name: str) -> str:
    return f"{dataset_name}.basic"


def generate_basic_suites(
    kedro_context: KedroContext,
    ge_context: DataContext,
    empty=False,
    replace=False,
    batch_kwargs=None,
):
    from great_expectations.profile import BasicSuiteBuilderProfiler

    if batch_kwargs is None:
        batch_kwargs = {}
    catalog = kedro_context.catalog

    existing_datasource_names = {ds["name"] for ds in ge_context.list_datasources()}
    for dataset_name in catalog.list():
        suite_name = generate_basic_suite_name(dataset_name)
        if suite_name in ge_context.list_expectation_suite_names() and not replace:
            continue

        datasource_name = generate_datasource_name(dataset_name)
        if datasource_name not in existing_datasource_names:
            continue

        dataset = catalog._get_dataset(dataset_name)
        data_path = str(dataset._filepath)
        dataasset_name, _ = os.path.splitext(os.path.basename(data_path))

        suite_batch_kwargs = {
            "datasource": datasource_name,
            "data_asset_name": dataasset_name,
            "path": data_path,
            "reader_options": dataset._load_args,
        }

        batch_kwargs_generator_name = "path"
        profiler_configuration = "demo"
        additional_batch_kwargs = batch_kwargs

        run_id = datetime.datetime.now(datetime.timezone.utc).strftime(
            "%Y%m%dT%H%M%S.%fZ"
        )

        if empty:
            create_empty_suite(ge_context, suite_name, suite_batch_kwargs)
        else:
            ge_context.profile_data_asset(
                datasource_name,
                batch_kwargs_generator_name=batch_kwargs_generator_name,
                data_asset_name=dataasset_name,
                batch_kwargs=suite_batch_kwargs,
                profiler=BasicSuiteBuilderProfiler,
                profiler_configuration=profiler_configuration,
                expectation_suite_name=suite_name,
                run_id=run_id,
                additional_batch_kwargs=additional_batch_kwargs,
            )



import datetime
import logging
from copy import copy
from typing import Any, Dict, List, Optional

import great_expectations as ge
from great_expectations.core.batch import Batch
from great_expectations.core.id_dict import BatchKwargs
from great_expectations.datasource.types import BatchMarkers
from great_expectations.validator.validator import Validator
from great_expectations.exceptions import ConfigNotFoundError
from kedro.framework.hooks import hook_impl
from kedro.io import DataCatalog

from .exceptions import UnsupportedDataSet
from .data import (
    get_suite_names,
    generate_datasource_name,
)


class KedroGreat:
    DEFAULT_SUITE_TYPES = ["warning", "basic", None]

    def __init__(
        self,
        expectations_map: Dict = None,
        suite_types: List[Optional[str]] = None,
        run_before_node: bool = True,
        run_after_node: bool = False,
    ):
        if expectations_map is None:
            expectations_map = {}
        if suite_types is None:
            suite_types = copy(KedroGreat.DEFAULT_SUITE_TYPES)
        self.expectations_map = expectations_map
        self.suite_types = suite_types

        self._before_node_run = run_before_node
        self._after_node_run = run_after_node
        self.logger = logging.getLogger("KedroGreat")
        self._finished_suites = set()

        try:
            self.expectation_context = ge.data_context.DataContext()
            self.expectation_suite_names = set(
                self.expectation_context.list_expectation_suite_names()
            )
        except ConfigNotFoundError:
            self.logger.error(
                "Great Expectations has not been initialized. "
                "KedroGreat cannot operate. "
                "Please run 'kedro great init'."
            )
            self.expectation_context = None

    @hook_impl
    def before_node_run(
        self, catalog: DataCatalog, inputs: Dict[str, Any], run_id: str
    ) -> None:
        if self._before_node_run:
            self._run_validation(catalog, inputs, run_id)

    @hook_impl
    def after_node_run(
        self, catalog: DataCatalog, outputs: Dict[str, Any], run_id: str
    ) -> None:
        if self._after_node_run:
            self._run_validation(catalog, outputs, run_id)

    def _run_validation(self, catalog: DataCatalog, data: Dict[str, Any], run_id: str):
        if self.expectation_context is None:
            return

        for dataset_name, dataset_value in data.items():
            ran_suite_for_dataset = False

            target_suite_names = get_suite_names(
                self.expectations_map, dataset_name, self.suite_types
            )

            dataset = catalog._get_dataset(dataset_name)
            dataset_path = str(dataset._filepath)
            df = dataset.load()

            try:
                for target_suite_name in target_suite_names:
                    if (
                        target_suite_name not in self.expectation_suite_names
                        or target_suite_name in self._finished_suites
                    ):
                        continue

                    self._run_suite(
                        dataset_name, dataset_path, df, target_suite_name, run_id
                    )

                    self._finished_suites.add(target_suite_name)
                    ran_suite_for_dataset = True

                if not ran_suite_for_dataset:
                    self.logger.warning(
                        f"Missing Expectation Suite for DataSet: {dataset_name}"
                    )
            except UnsupportedDataSet:
                self.logger.warning(
                    f"Unsupported DataSet Type: {dataset_name}({type(dataset)})"
                )

    def _run_suite(
        self,
        dataset_name: str,
        dataset_path: str,
        df: Any,
        target_expectation_suite_name: str,
        run_id: str,
    ):
        target_suite = self.expectation_context.get_expectation_suite(
            target_expectation_suite_name
        )
        batch_markers = BatchMarkers(
            {
                "ge_load_time": datetime.datetime.now(datetime.timezone.utc).strftime(
                    "%Y%m%dT%H%M%S.%fZ"
                )
            }
        )
        batch = Batch(
            "kedro",
            batch_kwargs=BatchKwargs(
                {
                    "path": dataset_path,
                    "datasource": generate_datasource_name(dataset_name),
                }
            ),
            data=df,
            batch_parameters=None,
            batch_markers=batch_markers,
            data_context=self.expectation_context,
        )

        try:
            v = Validator(batch=batch, expectation_suite=target_suite,)
        except ValueError:
            raise UnsupportedDataSet

        validator_dataset_batch = v.get_dataset()
        return self.expectation_context.run_validation_operator(
            "action_list_operator", [validator_dataset_batch], run_id=run_id
        )
